#####################################################################
# insertLemmas.py: Script to insert lemma information in PPCME2     #
# Author: Michael Percillier                                        #
# Date started: 2016-01-07                                          #
# Log: 2016-01-13: Script works as expected                         #
#      2016-01-14: Added information on verbs with French etymology #
#                  Added summary file output                        #
#                  Wrote read-me file --> v1                        #
#      2016-02-05: Implemented spelling variants for forms with no  #
#                  matching lemma                                   #
#      2016-02-09: Switch verb form to lowercase, to prevent        #
#                  cases of no-match for verb form in all-caps or   #
#                  at the beginning of sentences                    #
#                  Removed handling of MD (modal verbs)             #
#                  Changed () in MED lemmas to [], as this was      #
#                  problematic in CorpusSearch                      #
#      2016-03-31: Changed script to handle V tags only             #
#                  Changed script to deal with split verbs, e.g.    #
#             (VBP21 bi@l=NA@m=NA) (VBP22 tacne+d@l=toknen@m=46220) #
#                  is now treated as                                #
#             (VBP21 bi) (VBP22 tacne+d@l=bitoknen@m=4955)          #
#                  Added handling of verb forms preceded by $,      #
#                  e.g. $dreden                                     #
#      2016-04-05: Added use of the stem of unrecognised forms to   #
#                  match stems of recognised forms                  #
#                  Updated read-me file --> v4                      #
#      2016-05-27: Changed lemma information format to add @ at the #
#                  very end, so a search for MED-ID 15 doesn't      #
#                  return 150, 151, 1500 etc.                       #
#                  Included spelling substitution and stemming      #
#                  stats in summary file                            #
#                  Changed format of summary file to CSV that can   #
#                  be analysed as is in R                           #
#                  Updated (but not yet final) version of           #
#                  lemmas.yaml                                      #
#      2016-06-06: Updated to latest version of lemmas.yaml         #
#                  Added more spelling variation rules --> v5       #
#      2016-07-01: Update to final version of lemmas.yaml --> v6    #
#      2017-01-12: Insertion of e=nonfrench for verbs not part of   #
#                  list of French verbs provide by the OED --> v7   #
#      2017-07-28: Updated list of French-based verbs and version   #
#                  of lemmas.yaml --> v8                            #
#      2017-10-17: Updated list of French-based verbs and version   #
#                  of lemmas.yaml --> v9                            #
#      2018-02-05: Included most recent list of form-lemma links    #
#                  and French-based verbs                           #
#                  Summary option is now activated via an argument  #
#                  call (-s or --summary) rather than uncommenting  #
#                  specific lines of code                           #
#                  Merged version numbering with versions published #
#                  on the BASICS Toolkit website --> v2             #
#      2018-04-16: Included an updated (partially disambiguated and #
#                  manually corrected) list of form-lemma links     #
#                  Resulting file names are now in a new format that#
#                  begins with the manuscript date                  #
#      2018-09-27: Changed the format of the form-lemma list from   #
#                  YAML to CSV, as this is more transparent and     #
#                  doesn't require users to install the 'pyyaml'    #
#                  module in order to use the script                #
#                  --> v6                                           #
#      2018-10-08: Updated `lemmas.csv` --> v7                      #
#      2019-01-31: Updated `lemmas.csv` and `frenchMEDverbs.csv`,   #
#                  added treatment of `~` (tilde) character --> v8  #
#      2019-04-30: Minor fixes to make the script compatible with   #
#                  other corpora, in this case the PCMEP            #
#      2019-06-06: Minor fixes to handle word split and superscript #
#                  markers in PLAEME --> v9                         #
#      2019-10-22: Updated `lemmas.csv` to include forms that were  #
#                  not lemmatised in PCMEP and PLAEME               #
#                  Fixed an error that prevented PLAEME forms       #
#                  containing a ^ superscript marker from having    #
#                  their lemma information inserted --> v10         #
#      2019-11-05: Minor corrections and additions to `lemmas.csv`  #
#                  --> v11                                          #
#      2020-07-17: Minor corrections and additions to `lemmas.csv`, #
#                  some comments are now marked as warnings,        #
#                  rather than an additional lemma suggestion       #
#                  --> v12                                          #
#      2020-09-15: Added insertion of a disclaimer in the output    #
#                  files so that end users are made aware of the    #
#                  fact that their corpus files have been treated   #
#                  by the script, with links to further information #
#                  Fixed a minor compatibility issue between        #
#                  Python 2 and Python 3 related to the display of  #
#                  credits when running the script                  #
#                  --> v13                                          #
#####################################################################

# Load required modules
import re # Regular expressions
import sre_constants # Handle Regex errors
import codecs # Unicode support
import os # Access/create directories
import sys # Needed to display user progress
import argparse # To specify input folder path directly in the command line
import collections# For OrderedDict (ordered dictionary/hash)
from itertools import product # To generate permutations based on spelling variants

# Set version number
versionNumber = 13

# Set disclaimer
disclaimer = """
/*
DISCLAIMER:
This corpus file was treated by a script to lemmatize verb forms: insertLemmas.py, version %d, created by Michael Percillier within the research project Borrowing of Argument Structure in Contact Situations (BASICS).

Each piece of inserted information is demarcated by @ characters and specified by an attribute, such as:
- l: the verb lemma as specified in the Middle English Dictionary (MED)
- m: the MED-ID, i.e. the numerical identifier used in the MED
- e: etymological information, which at the moment only distinguishes French-based verbs and therefore has the possible values "french" and "nonfrench"
- w: warnings, either marking that annotators had doubts about the appropriate lemma, or that spelling substitution or stemming methods were used to estimate a lemma

For more information on the lemmatizing script, please consult the page "Lemmatizer" on the BASICS Toolkit website http://www.tinyurl.com/basicstoolkit and the publications mentioned therein.

For more information on the BASICS research project, please consult the pages https://tinyurl.com/dfgbasics and https://gepris.dfg.de/gepris/projekt/265711632
*/

""" % versionNumber

# define dictionary of substitutions
substitutions = {
	"i": ["e","y"],
	"e": ["i"],
	"y": ["i","g","+g"],
	"u": ["v","ou"],
	"v": ["u"],
	"th": ["+t","+d"],
	"+t": ["th"],
	"+d": ["th"],
	"g": ["+g","y"],
	"+g": ["g","y"],
	"ou": ["u"],
	"ll": ["l"],
	"nn": ["n"],
	"pp": ["p"],
	"-": [""],
	"~": [""]
}
# Added ou -> u, u -> ou, ll -> l, nn -> n, pp -> p, - -> "", add g, +g to y, add y to g and +g


# define function to produce spelling variants
def variants(form):
	# because of th, +t, +d, +g, a character-based approach can't be used
	# cut the form into chunks that take these combinations into account
	# Not the most elegant way to chunks the forms, but it seems to work:
	# put spaces around defined character sequences, and all other single characters:
	chunks = re.sub("(\+t|\+d|\+g|th|.)"," \\1 ",form)
	# and split by spaces
	chunks = chunks.split()
	possibilities = []
	# go through each chunk
	for chunk in chunks:
		# observed chunk as first item
		options = [chunk]
		# add alternate chunks if applicable
		if chunk in list(substitutions.keys()):
			for item in substitutions[chunk]:
				options.append(item)
		possibilities.append(options)
	# use the product() function from itertools module to iterate through all possible combinations
	possibleVariants = []
	for possibility in product(*possibilities):
		possibleVariants.append("".join(possibility))
	return possibleVariants

# define function to pre-process PLAEME forms
def plaemeProcessing(verbForm):
	processedForm = re.sub(r"\+c","c",verbForm) # remove any c-ligature marking
	processedForm = re.sub("(g|G)2","\\1",processedForm) # replace alternative 'yogh' with regular 'yogh'
	processedForm = re.sub(r"\+(w|W)","\\1",processedForm) # replace 'wynn' with 'w'
	processedForm = re.sub("'","",processedForm) # remove accents
	processedForm = re.sub("=","",processedForm) # remove morpheme boundary
	processedForm = re.sub("@","",processedForm) # remove word split marker
	processedForm = re.sub(r"\^","",processedForm) # remove superscript marker
	processedForm = re.sub(r"^(.*)\-.*?$", "\\1",processedForm) # remove lexel
	return processedForm

# define function to check matching lemmas to a verb form
def checkLemmas(verbForm):
	global multipleCounter
	lowerVerbForm = verbForm.lower() # use a lowercase version of the observed form, which may be in all-caps or at the beginning of a sentence
	lemmaMatches = [] # list to store the lemmas matching the form
	for lemma in list(hash.keys()): # check each lemma whether it has a matching form
		if lowerVerbForm in list(hash[lemma].keys()):
			lemmaMatches.append(lemma)
			# include doubt markers and additional lemma suggestions
			if len(hash[lemma][lowerVerbForm]) > 0: # an empty list means there are no comments or additional lemmas to collect
				multipleCounter += 1
				for additionalItem in hash[lemma][lowerVerbForm]: # no need for frequency information
					lemmaMatches.append(additionalItem)
	return lemmaMatches

# define function to handle matching lemmas
def handleMatches(verbForm,lemmaMatches):
	global frenchCounter
	lemmaInfo = verbForm # the verb form will be replaced by itself followed by lemma attributes
	for lemmaMatch in lemmaMatches:
		try:
			# properly formed lemma entries with an MED ID at the end
			lemmaInfo = lemmaInfo + "@l=%s@m=%s" % (lemmaMatch.split("_MED")[0],lemmaMatch.split("_MED")[1])
			# check if the MED lemma if part of the French verbs provided by the OED
			# and if so, mark the etymology as French
			if lemmaMatch.split("_MED")[1] in frenchMEDids:
				frenchCounter += 1
				lemmaInfo = lemmaInfo + "@e=french"
			# or as non-French if not	
			else:
				lemmaInfo = lemmaInfo + "@e=nonfrench"
		except IndexError:
			# take care of malformed/missing MED ID
			lemmaInfo = lemmaInfo + "@l=%s@m=NA" % lemmaMatch
	return lemmaInfo	

# define function to stem verb forms
def stemmer(verbForm,stemList,suffixes):
	for suffix in suffixes:
		lowerVerbForm = verbForm.lower()
		stem = re.sub(suffix+"$","",lowerVerbForm) # $ because character string should only be removed from the end
		if lowerVerbForm != stem: # only save result if it is different from the input form
			if stem not in stemList:
				stemList.append(stem)
	return stemList


# save ANSI escape sequences to make messages to user pop out
ANSIreverse = "\033[7m"
ANSIreset = "\033[0m"

# Info banner
print("\n" + ANSIreverse)
print("{:^80}".format("insertLemmas.py: Add lemma information to verbs in PPCME2, PCMEP, or PLAEME"))
print("{:^80}".format("Created by Michael Percillier, University of Mannheim, 2016-2020"))
print("{:^80}".format("Version %d" % versionNumber))
print(ANSIreset + "\n")

# Read arguments in program call
try:
	parser = argparse.ArgumentParser()
	parser.add_argument("directory",help="name of the corpus directory")
	parser.add_argument("-s", "--summary", help="Generate a summary of lemmatised verbs",action="store_true")
	parser.add_argument("-p", "--plaeme", help="Indicate whether PLAEME texts are to be processed",action="store_true")
	parser.add_argument("-d", "--dating", help="Indicate whether you want manuscript and composition dates to be placed in the file names (PPCME2 only)",action="store_true")
	args = parser.parse_args()
	corpusDirName = args.directory
except SystemExit: # if user didn't specify directory name in program call, ask interactively
	print("\n"+ANSIreverse+"You haven't specified a corpus folder in the program call." + ANSIreset)
	corpusDirName = input("Please enter the name of the corpus folder:")

# Open folder containing corpus
fileList = os.listdir(os.getcwd() + "/" +  corpusDirName)

# Create output directory to contain lemmatised versions of corpus files
# Name is the name of the input folder, with _Lemmatised appended
# If the output directory already exists, ask user to either overwrite it or quit.
try:
	os.mkdir(os.getcwd() + "/" +  corpusDirName + "_Lemmatised")
except OSError:
	print(ANSIreverse+"Intended output folder already exists.\nDo you wish to proceed?"+ANSIreset)
	overwriteAnswer = ""
	while not re.match("[yYnN]",overwriteAnswer):
		overwriteAnswer = input("(Y)es [overwrite]\t(N)o [quit] ")
	if re.match("[nN]",overwriteAnswer):
		print("Quitting program. Delete or rename the folder *_Lemmatised and rerun the script.")
		raise SystemExit

# Import inventory of verb form/lemma matches
# NB: The inventory is now in CSV rather than in the YAML format,
#     based on the spreadsheets made by the student assistants
print("\nLoading lemmas ...") # Inform user (as it can take a few moments)
hash = {}
CSVfile = open("lemmas.csv", "r")
for line in CSVfile:
	if re.match("Form;Lemma",line):
		pass #Ignore header line
	else:
		try:
			currentForm = line.split(";")[0].strip().lower()
			currentLemma = line.split(";")[1].strip()
			currentComment = line.split(";")[2].strip()
			currentSuggestions = line.split(";")[3].strip()
			currentSuggestions = currentSuggestions.split("|")
		except IndexError:
			pass # last empty line should be ignored
		if currentLemma in list(hash.keys()):
			if currentForm in list(hash[currentLemma].keys()):
				if currentComment not in hash[currentLemma][currentForm]:
					if currentComment != "":
						[currentLemma][currentForm].append(currentComment)
				for suggestion in currentSuggestions:
					if suggestion not in hash[currentLemma][currentForm]:
						if suggestion != "":
							hash[currentLemma][currentForm].append(suggestion)
			else:
				hash[currentLemma][currentForm] = []
				if currentComment != "":
					hash[currentLemma][currentForm].append(currentComment)
				for suggestion in currentSuggestions:
					if suggestion != "":
						hash[currentLemma][currentForm].append(suggestion)
		else:
			hash[currentLemma] = {}
			hash[currentLemma][currentForm] = []
			if currentComment != "":
				hash[currentLemma][currentForm].append(currentComment)
			for suggestion in currentSuggestions:
				if suggestion != "":
					hash[currentLemma][currentForm].append(suggestion)
CSVfile.close()
print("Lemmas loaded.\n")


# make a stemHash containing stemmed versions of all form in hash
# same keys as in hash, but contents are stemmed lemmas + stemmed verb forms
# available stemmers don't work for Middle English, so make a rudimentary one
# define a set of ME suffixes that remove from the end of verbs to obtain stems
MEsuffixes = [
"\+d",
"\+d\+d",
"\+t",
"\+t\+t",
"an",
"ande?",
"dd?",
"den?",
"e",
"e\+d",
"e\+t",
"ede?",
"enn?",
"e?st",
"et",
"in?d?e?",
"ingg?e?",
"ode",
"odest",
"oden",
"ten?",
"th",
"tt?",
"yde?",
"ynde?",
"ynn?",
"yngg?e?",
"yst"]

# save stemmed forms of each lemma in a new hash
stemHash = {}
print("\nStemming forms ...") # Inform user
for lemma in list(hash.keys()):
	currentStems = [] # stores created stems for current lemma
	# stem lemma first
	currentLemma = lemma.split("_")[0] # isolate form from MED lemma number
	currentStems = stemmer(currentLemma,currentStems,MEsuffixes)
	# stem forms listed as matching the lemma
	types = list(hash[lemma].keys())
	for type in types:
		currentStems = stemmer(type,currentStems,MEsuffixes)
	# add forms in stemHash under lemma ID corresponding to main hash
	stemHash[lemma] = currentStems
print("Forms stemmed.\n") # Inform user

# Import list of MED IDs linked to the OED entries of verbs of French origin provided by the OED
MEDfrench = codecs.open("frenchMEDverbs.csv",encoding="utf-8",mode="r")
frenchMEDids = [] # list to hold MED IDs
for line in MEDfrench:
	currentMEDid = line.split(";")[2]
	if re.match("\d",currentMEDid):
		frenchMEDids.append(currentMEDid)
MEDfrench.close()

# Import text dates to prepend dates to output files
textDates = codecs.open("textDates.csv",encoding="utf-8",mode="r")
textDateHash = {}
for line in textDates:
	textName = line.split(";")[0]
	msDate = line.split(";")[1]
	compoDate = line.split(";")[2].strip()
	textDateHash[textName] = [msDate,compoDate]

print ("Processing corpus files ...")

fileCounter = 0 # used later to inform user of progress

stats = collections.OrderedDict() # hash to store info, to make a summary per file (and a global one)

# go through each corpus file
for corpusFile in fileList:
	if re.search("\.psd$",corpusFile):# only treat files ending in .psd, exclude any system files such as .DS_Store et al.
		# create corresponding output file in output folder
		# file should be the same as the original, but a .lemma before the .psd ending
		# also, manuscript and composition date should be prepended
		# to do so, determine text code
		if args.dating:# for PPCME2: add manuscript and composition dates in file name
			textCode = re.findall("cm.+?\.",corpusFile)[0].strip(".")
			# one text (cmancriw) is split in two with the suffixes -1 and -2, so remove these
			textCode = re.sub("\-\d$","",textCode)
			# define the output file name as date[m]-date[c]-originalName.lemma.psd
			outpFileName = textDateHash[textCode][0] + "[m]-" + textDateHash[textCode][1] + "[c]-" + re.sub("\.psd",".lemma.psd",corpusFile)
		else:# generic handling of corpora other than PPCME2: just add ".lemma" before ".psd" file suffix
			outpFileName = re.sub("\.psd",".lemma.psd",corpusFile)
		outpFilePath = os.getcwd() + "/" +  corpusDirName + "_Lemmatised"+"/" + outpFileName
		outp = codecs.open(outpFilePath,encoding="utf-8",mode="w")
		# Insert disclaimer
		outp.write(disclaimer)
		# open the input file
		inp = codecs.open(os.getcwd() + "/" + corpusDirName + "/" + corpusFile,encoding="utf-8",mode="r")
		# make verb counters to get stats such as
		#     - total number of verbs
		#     - verbs matched directly
		#     - verbs with no matching lemma
		#     - verbs of French origin
		#     - verbs with multiple suggestions and/or doubt markers
		#     - verbs that were matched using spelling substitution
		#     - verbs that were matched by stemming
		totalVerbCounter = 0
		directMatchCounter = 0
		noMatchCounter = 0
		frenchCounter = 0
		multipleCounter = 0
		subsCounter = 0
		stemCounter = 0
		# go through each line of the input file
		for line in inp:
			# check if line contains a V-tagged verb:
			if re.search("\(V",line):
				lineVerbs = re.findall("\(V\S+ \$?[^)]+",line) #isolate verb(s) in line (NB: some verb forms are preceded by a dollar sign)
				# Explanation: (TAG as above followed by a space and 1 or more word-characters with possible +-~ characters in between
				updatedLine = line # make a copy of the line to be modified
				for verb in lineVerbs:# handle the verb(s) in the line
					# handle split verbs, which should only count as one but have V21, V22-type tags
					# check if the current form is part of a split verb
					splitIndex = re.findall("\(V\w+\d+",verb) # get tag with index digits
					try:
						splitIndex = splitIndex[0]
						splitIndex = re.sub("\D","",splitIndex) # keep digits only
					except IndexError:
						splitIndex = "00" # special for unsplit forms
					
					# the splitIndex contains 2 digits: 1st the number of items, 2nd the index of the current item
					# if both digits are identical, then we are dealing with the last item of the split verb
					# if the digits are different, the final item is not yet reached
					if splitIndex[0] != splitIndex[1]: # deal with intermediate items first
						if splitIndex[1] == "1":
							lookUpForm = re.sub("\(V\S+ \$?","",verb)
						else:
							lookUpForm = lookUpForm + re.sub("\(V\S+ \$?","",verb)
						# for intermediate split items, the treatment of the form stops here
					else: # for final items of split forms and unsplit forms, a matching lemma is queried
						if splitIndex == "00": # if the verb form is not split, verbForm and lookUpForm are identical
							verbForm = re.sub("\(V\S+ \$?","",verb)
							lookUpForm = verbForm 
						else: # handle last item of split verb
							verbForm = re.sub("\(V\S+ \$?","",verb) # form after which the lemma info will be placed
							lookUpForm = lookUpForm + verbForm # form used to check for lemmas
						totalVerbCounter += 1
						# handle all V*, and match forms to lemma suggestions
						# in case texts from PLAEME are being processed, the form needs some pre-processing first
						if args.plaeme:
							lookUpForm = plaemeProcessing(lookUpForm)
						# first check verb form as is for lemma matches
						plainMatches = checkLemmas(lookUpForm)
						# handle found matches
						if len(plainMatches) > 0:
							lemmaInfo = handleMatches(verbForm,plainMatches)
							directMatchCounter += 1
						# if there are no matches, apply some character substitutions and check for lemmas again
						else:
							lemmaInfo = verbForm
							alternateForms = variants(lookUpForm)
							for alternateForm in alternateForms:
								if alternateForm != lookUpForm:# only substitutions that led to a change
									alternateMatches = checkLemmas(alternateForm)
									if len(alternateMatches) > 0:
										lemmaInfo = handleMatches(lemmaInfo,alternateMatches)
										# add a marker signalling that the match is based on a substitution
										lemmaInfo = lemmaInfo + "@w=substitution"
										subsCounter += 1
										# do not continue once a variant form has matched a lemma,
										# otherwise, the same lemma information may be repeated over and over at each subsequent match:
										break
							# if there are no matches after applying substitutions, use stemming and check for matches
							# compare lists with set(list1).intersection(list2) ?
							# returns set([]) if no matches in common, so use != set([]) to see of any matches were found
							if lemmaInfo == verbForm: # lemmaInfo hasn't been modified if substitution didn't return any hits
								stemMatches = [] # store lemmas whose stemmed forms matched those of the token
								# generate stems for the token in the corpus
								tokenStems = []
								tokenStems = stemmer(lookUpForm,tokenStems,MEsuffixes)
								# and compare them with the stems of lemmatised forms
								for stemLemma in list(stemHash.keys()):
									stemsComparison = set(tokenStems).intersection(stemHash[stemLemma])
									if stemsComparison != set([]):
										stemMatches.append(stemLemma)
								if len(stemMatches) > 0:
									lemmaInfo = handleMatches(verbForm,stemMatches)
									# add a marker signalling that the match is based on stemming
									lemmaInfo = lemmaInfo + "@w=stemming"
									stemCounter += 1
							# if none of these alternative forms returned matches, add NA as attributes
							if lemmaInfo == verbForm: # lemmaInfo hasn't changed from its original value if none of the forms returned matches
								noMatchCounter += 1
								# mark forms with no matching lemmas
								lemmaInfo = verbForm + "@l=NA@m=NA"
					# now that all cases are taken care of, update line to include lemma information
					
					# some MED lemmas have letters in brackets to mark spelling variants, e.g. assail(l)en
					# such brackets are problematic for CorpusSearch, as brackets mark nodes in the Penn-format
					# replace round brackets with square brackets
					lemmaInfo = re.sub("\(","[",lemmaInfo)
					lemmaInfo = re.sub("\)","]",lemmaInfo)
					# remove any whitespace that may have been introduced
					lemmaInfo = re.sub("\s+","",lemmaInfo)
					# add a @ at the end to "close" any MED-ID at the end;
					# otherwise, as search for m=15 can yield 150
					# by adding @, the search can be specified as m=15@
					lemmaInfo = lemmaInfo + "@"
					# any + and ^ in the verbForm need to be escaped to update the line
					escapedVerbForm = re.sub("\+","\+",verbForm)
					escapedVerbForm = re.sub("\^","\^", escapedVerbForm)
					try:
						updatedLine = re.sub(escapedVerbForm,lemmaInfo,updatedLine)
						#replace the following comments marked as @l= with @w=
						# to mark them as warnings and not as lemma suggestions:
						# - doubt!* (sometimes has an exclamation mark)
						# - noun
						# - adjective (cases of wrong POS in the corpus tagging)
						# - compound (lemmatised verb occurs as part of a compound)
						# - wrong-lexel\?* (disagreement with the proposed lexel in PLAEME)
						updatedLine = re.sub("@l=(doubt!*|noun|adjective|compound|wrong-lexel\?*)@m=NA","@w=\\1",updatedLine)
					except sre_constants.error:# error raised in case of traces, i.e. when the form is "*", in which case the line should be unchanged
						updatedLine = line
				# write updated line to output
				outp.write(updatedLine)
			else: # in case no verb was found in the line, copy the input line as is
				outp.write(line)
		# store file stats in stats hash
		stats[corpusFile] = collections.OrderedDict()
		stats[corpusFile]["VerbsTotal"] = totalVerbCounter
		stats[corpusFile]["DirectMatches"] = directMatchCounter
		stats[corpusFile]["SubstitutionMatches"] = subsCounter
		stats[corpusFile]["StemmingMatches"] = stemCounter
		stats[corpusFile]["NoMatches"] = noMatchCounter
		stats[corpusFile]["FrenchEtymology"] = frenchCounter
		stats[corpusFile]["MultipleSuggestionsOrDoubts"] = multipleCounter
		# close input and output files
		inp.close()
		outp.close()
	# Inform user of progress
	fileCounter += 1
	sys.stdout.write("\rProcessed file %d / %d " % (fileCounter,len(fileList)))
	sys.stdout.flush()
# Let user know that the entire process is complete
print("\nProcess complete!\n")
print()
print("Lemmatised corpus created in the following directory: %s" %  ANSIreverse+corpusDirName+"_Lemmatised"+ANSIreset)
print()

# Save stats for each corpus file in a summary file if specified by user in the program call
if args.summary:
	summary = codecs.open(corpusDirName+"_Lemmatised_Summary.csv",encoding="utf-8",mode="w")
	summary.write("File;Period;Base;VerbsTotal;DirectMatches;SubstitutionMatches;StemmingMatches;NoMatches;FrenchEtymology;MultipleSuggestionsOrDoubts\n") # header
	# One line (row) per file:
	for item in list(stats.keys()):
		summary.write(item)
		summary.write(";")
		# first 2 characters of file name carry period
		summary.write(item[0:2])
		summary.write(";")
		# fourth and fifth characters carry the textual basis (French-based text or original English text)
		summary.write(item[3:5])
		for count in list(stats[item].keys()):
			summary.write(";")
			summary.write(str(stats[item][count]))
		summary.write("\n")
	print("Summary saved as: %s" % ANSIreverse+corpusDirName+"_Lemmatised_Summary.csv"+ANSIreset)
	print()	
				
				